@@ -4,7 +4,7 @@
  * See Copyright for the status of this software.
  *
  * Daniel.Veillard@imag.fr
- * Bjorn Reese <breese@mail1.stofanet.dk> for number formatting
+ * Bjorn Reese <breese@users.sourceforge.net> for number formatting
  */
 
 #ifndef __XML_XSLT_FUNCTIONS_H__
@@ -18,25 +18,6 @@
 extern "C" {
 #endif
 
-/*
- * Data structure of decimal-format
- */
-typedef struct _xsltDecimalFormat {
-    /* Used for interpretation of pattern */
-    xmlChar digit;
-    xmlChar patternSeparator;
-    /* May appear in result */
-    xmlChar minusSign;
-    xmlChar *infinity;
-    xmlChar *noNumber;
-    /* Used for interpretation of pattern and may appear in result */
-    xmlChar decimalPoint;
-    xmlChar grouping;
-    xmlChar percent;
-    xmlChar permille;
-    xmlChar zeroDigit;
-} xsltDecimalFormat, *xsltDecimalFormatPtr;
-
 /*
  * Interfaces for the functions implementations
  */
