@@ -62,8 +62,6 @@ struct _xsltStepOp {
     xmlChar *value3;
 };
 
-typedef struct _xsltCompMatch xsltCompMatch;
-typedef xsltCompMatch *xsltCompMatchPtr;
 struct _xsltCompMatch {
     struct _xsltCompMatch *next; /* siblings in the name hash */
     float priority;              /* the priority */
@@ -94,15 +92,13 @@ struct _xsltParserContext {
 
 /**
  * xsltNewCompMatch:
- * @mode:  the mode name or NULL
- * @modeURI:  the mode URI or NULL
  *
  * Create a new XSLT CompMatch
  *
  * Returns the newly allocated xsltCompMatchPtr or NULL in case of error
  */
 xsltCompMatchPtr
-xsltNewCompMatch(const xmlChar *mode, const xmlChar *modeURI) {
+xsltNewCompMatch(void) {
     xsltCompMatchPtr cur;
 
     cur = (xsltCompMatchPtr) xmlMalloc(sizeof(xsltCompMatch));
@@ -113,8 +109,6 @@ xsltNewCompMatch(const xmlChar *mode, const xmlChar *modeURI) {
     }
     memset(cur, 0, sizeof(xsltCompMatch));
     cur->maxStep = 20;
-    cur->mode = xmlStrdup(mode);
-    cur->modeURI = xmlStrdup(modeURI);
     return(cur);
 }
 
@@ -575,6 +569,32 @@ xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
     return(1);
 }
 
+/**
+ * xsltTestCompMatchList:
+ * @ctxt:  a XSLT process context
+ * @node: a node
+ * @comp: the precompiled pattern list
+ *
+ * Test wether the node matches one of the patterns in the list
+ *
+ * Returns 1 if it matches, 0 if it doesn't and -1 in case of failure
+ */
+int
+xsltTestCompMatchList(xsltTransformContextPtr ctxt, xmlNodePtr node,
+	              xsltCompMatchPtr comp) {
+    int ret;
+
+    if ((ctxt == NULL) || (node == NULL))
+	return(-1);
+    while (comp != NULL) {
+	ret = xsltTestCompMatch(ctxt, comp, node, NULL, NULL);
+	if (ret == 1)
+	    return(1);
+	comp = comp->next;
+    }
+    return(0);
+}
+
 /************************************************************************
  *									*
  *			Dedicated parser for templates			*
@@ -1096,102 +1116,119 @@ error:
 /**
  * xsltCompilePattern:
  * @pattern an XSLT pattern
- * @mode:  the mode name or NULL
- * @modeURI:  the mode URI or NULL
  *
- * Compile the XSLT pattern and generates a precompiled form suitable
+ * Compile the XSLT pattern and generates a list of precompiled form suitable
  * for fast matching.
- * Note that the splitting as union of patterns is expected to be handled
- * by the caller
  *
  * [1] Pattern ::= LocationPathPattern | Pattern '|' LocationPathPattern
  *
- * Returns the generated xsltCompMatchPtr or NULL in case of failure
+ * Returns the generated pattern list or NULL in case of failure
  */
 
 xsltCompMatchPtr
-xsltCompilePattern(const xmlChar *pattern, const xmlChar *mode,
-	           const xmlChar *modeURI) {
-    xsltParserContextPtr ctxt;
-    xsltCompMatchPtr ret;
-    const xmlChar *cur;
+xsltCompilePattern(const xmlChar *pattern) {
+    xsltParserContextPtr ctxt = NULL;
+    xsltCompMatchPtr element, first = NULL, previous = NULL;
+    int current, start, end;
 
     if (pattern == NULL) {
         xsltGenericError(xsltGenericErrorContext,
-		"xsltCompilePattern : NULL pattern\n");
+			 "xsltCompilePattern : NULL pattern\n");
 	return(NULL);
     }
 
 #ifdef DEBUG_PARSING
     xsltGenericDebug(xsltGenericDebugContext,
-	    "xsltCompilePattern : parsing '%s'\n", pattern);
+		     "xsltCompilePattern : parsing '%s'\n", pattern);
 #endif
 
-    cur = pattern;
-    while (IS_BLANK(*cur)) cur++;
-    if (*cur == 0) {
-        xsltGenericError(xsltGenericErrorContext,
-		"xsltCompilePattern : NULL pattern\n");
-	return(NULL);
-    }
     ctxt = xsltNewParserContext();
     if (ctxt == NULL)
 	return(NULL);
-    ret = xsltNewCompMatch(mode, modeURI);
-    if (ret == NULL) {
-	xsltFreeParserContext(ctxt);
-	return(NULL);
+    current = end = 0;
+    while (pattern[current] != 0) {
+	start = current;
+	while (IS_BLANK(pattern[current]))
+	    current++;
+	end = current;
+	while ((pattern[end] != 0) && (pattern[end] != '|'))
+	    end++;
+	if (current == end) {
+	    xsltGenericError(xsltGenericErrorContext,
+			     "xsltCompilePattern : NULL pattern\n");
+	    goto error;
+	}
+	element = xsltNewCompMatch();
+	if (element == NULL) {
+	    goto error;
+	}
+	if (first == NULL)
+	    first = element;
+	else if (previous != NULL)
+	    previous->next = element;
+	previous = element;
+
+	ctxt->comp = element;
+	ctxt->base = xmlStrndup(&pattern[start], end - start);
+	ctxt->cur = &(ctxt->base)[current - start];
+	xsltCompileLocationPathPattern(ctxt);
+	if (ctxt->base)
+	    xmlFree((xmlChar *)ctxt->base);
+	if (ctxt->error)
+	    goto error;
+
+	/*
+	 * Reverse for faster interpretation.
+	 */
+	xsltReverseCompMatch(element);
+
+	/*
+	 * Set-up the priority
+	 */
+	if (((element->steps[0].op == XSLT_OP_ELEM) ||
+	     (element->steps[0].op == XSLT_OP_ATTR)) &&
+	    (element->steps[0].value != NULL) &&
+	    (element->steps[1].op == XSLT_OP_END)) {
+	    element->priority = 0;
+	} else if ((element->steps[0].op == XSLT_OP_PI) &&
+		   (element->steps[0].value != NULL) &&
+		   (element->steps[1].op == XSLT_OP_END)) {
+	    element->priority = 0;
+	} else if ((element->steps[0].op == XSLT_OP_NS) &&
+		   (element->steps[0].value != NULL) &&
+		   (element->steps[1].op == XSLT_OP_END)) {
+	    element->priority = -0.25;
+	} else if (((element->steps[0].op == XSLT_OP_PI) ||
+		    (element->steps[0].op == XSLT_OP_TEXT) ||
+		    (element->steps[0].op == XSLT_OP_ALL) ||
+		    (element->steps[0].op == XSLT_OP_NODE) ||
+		    (element->steps[0].op == XSLT_OP_COMMENT)) &&
+		   (element->steps[1].op == XSLT_OP_END)) {
+	    element->priority = -0.5;
+	} else {
+	    element->priority = 0.5;
+	}
+	if (pattern[end] == '|')
+	    end++;
+	current = end;
     }
-
-    ctxt->comp = ret;
-    ctxt->base = pattern;
-    ctxt->cur = cur;
-    xsltCompileLocationPathPattern(ctxt);
-    if (ctxt->error)
+    if (end == 0) {
+        xsltGenericError(xsltGenericErrorContext,
+			 "xsltCompilePattern : NULL pattern\n");
 	goto error;
-
-    /*
-     * Reverse for faster interpretation.
-     */
-    xsltReverseCompMatch(ret);
-
-    /*
-     * Set-up the priority
-     */
-    if (((ret->steps[0].op == XSLT_OP_ELEM) ||
-	 (ret->steps[0].op == XSLT_OP_ATTR)) &&
-	(ret->steps[0].value != NULL) &&
-	(ret->steps[1].op == XSLT_OP_END)) {
-	ret->priority = 0;
-    } else if ((ret->steps[0].op == XSLT_OP_PI) &&
-	       (ret->steps[0].value != NULL) &&
-	       (ret->steps[1].op == XSLT_OP_END)) {
-	ret->priority = 0;
-    } else if ((ret->steps[0].op == XSLT_OP_NS) &&
-	       (ret->steps[0].value != NULL) &&
-	       (ret->steps[1].op == XSLT_OP_END)) {
-	ret->priority = -0.25;
-    } else if (((ret->steps[0].op == XSLT_OP_PI) ||
-		(ret->steps[0].op == XSLT_OP_TEXT) ||
-		(ret->steps[0].op == XSLT_OP_NODE) ||
-	        (ret->steps[0].op == XSLT_OP_COMMENT)) &&
-	       (ret->steps[1].op == XSLT_OP_END)) {
-	ret->priority = -0.5;
-    } else {
-	ret->priority = 0.5;
     }
 
     xsltFreeParserContext(ctxt);
-    return(ret);
+    return(first);
 
 error:
-    xsltFreeParserContext(ctxt);
-    xsltFreeCompMatch(ret);
+    if (ctxt != NULL)
+	xsltFreeParserContext(ctxt);
+    if (first != NULL)
+	xsltFreeCompMatchList(first);
     return(NULL);
-
 }
 
-
 /************************************************************************
  *									*
  *			Module interfaces				*
@@ -1212,45 +1249,31 @@ error:
 int
 xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
 	        const xmlChar *mode, const xmlChar *modeURI) {
-    xsltCompMatchPtr pat, list, *top = NULL;
+    xsltCompMatchPtr pat, list, *top = NULL, next;
     const xmlChar *name = NULL;
-    xmlChar *p, *pattern, tmp;
 
-    if ((style == NULL) || (cur == NULL))
+    if ((style == NULL) || (cur == NULL) || (cur->match == NULL))
 	return(-1);
 
-    p = cur->match;
-    if (p == NULL)
-	return(-1);
+    pat = xsltCompilePattern(cur->match);
+    while (pat) {
+	next = pat->next;
+	pat->next = NULL;
+	
+	pat->template = cur;
+	if (mode != NULL)
+	    pat->mode = xmlStrdup(mode);
+	if (modeURI != NULL)
+	    pat->modeURI = xmlStrdup(modeURI);
+	if (pat->priority != XSLT_PAT_NO_PRIORITY)
+	    cur->priority = pat->priority;
+	else
+	    pat->priority = cur->priority;
 
-next_pattern:
-    if (*p == 0)
-	return(0);
-    /*
-     * get a compiled form of the pattern
-     */
-    pattern = p;
-    while ((*p != 0) && (*p != '|')) {
-	/* TODO: handle string escaping "a | b" in patterns ... */
-	p++;
-    }
-
-    tmp = *p;
-    *p = 0;
-    pat = xsltCompilePattern(pattern, mode, modeURI);
-    *p = tmp;
-    if (tmp != 0)
-	p++;
-    if (pat == NULL)
-	return(-1);
-    pat->template = cur;
-    if (cur->priority != XSLT_PAT_NO_PRIORITY)
-	pat->priority = cur->priority;
-
-    /*
-     * insert it in the hash table list corresponding to its lookup name
-     */
-    switch (pat->steps[0].op) {
+	/*
+	 * insert it in the hash table list corresponding to its lookup name
+	 */
+	switch (pat->steps[0].op) {
         case XSLT_OP_ATTR:
 	    if (pat->steps[0].value != NULL)
 		name = pat->steps[0].value;
@@ -1262,29 +1285,29 @@ next_pattern:
         case XSLT_OP_PARENT:
         case XSLT_OP_ANCESTOR:
         case XSLT_OP_NS:
-             name = pat->steps[0].value;
-	     break;
+	    name = pat->steps[0].value;
+	    break;
         case XSLT_OP_ROOT:
-             top = (xsltCompMatchPtr *) &(style->rootMatch);
-	     break;
+	    top = (xsltCompMatchPtr *) &(style->rootMatch);
+	    break;
         case XSLT_OP_KEY:
-             top = (xsltCompMatchPtr *) &(style->keyMatch);
-	     break;
+	    top = (xsltCompMatchPtr *) &(style->keyMatch);
+	    break;
         case XSLT_OP_ID:
-	     /* TODO optimize ID !!! */
+	    /* TODO optimize ID !!! */
         case XSLT_OP_ALL:
-             top = (xsltCompMatchPtr *) &(style->elemMatch);
-	     break;
+	    top = (xsltCompMatchPtr *) &(style->elemMatch);
+	    break;
         case XSLT_OP_END:
 	case XSLT_OP_PREDICATE:
 	    xsltGenericError(xsltGenericErrorContext,
-		    "xsltAddTemplate: invalid compiled pattern\n");
+			     "xsltAddTemplate: invalid compiled pattern\n");
 	    xsltFreeCompMatch(pat);
 	    return(-1);
-	/*
-	 * TODO: some flags at the top level about type based patterns
-	 *       would be faster than inclusion in the hash table.
-	 */
+	    /*
+	     * TODO: some flags at the top level about type based patterns
+	     *       would be faster than inclusion in the hash table.
+	     */
 	case XSLT_OP_PI:
 	    if (pat->steps[0].value != NULL)
 		name = pat->steps[0].value;
@@ -1304,98 +1327,101 @@ next_pattern:
 		top = (xsltCompMatchPtr *) &(style->elemMatch);
 	    
 	    break;
-    }
-    if (name != NULL) {
-	if (style->templatesHash == NULL) {
-	    style->templatesHash = xmlHashCreate(1024);
+	}
+	if (name != NULL) {
 	    if (style->templatesHash == NULL) {
-		xsltFreeCompMatch(pat);
-		return(-1);
-	    }
-#ifdef DEBUG_PARSING
-	    xsltGenericDebug(xsltGenericDebugContext,
-		    "xsltAddTemplate: created template hash\n");
-#endif
-	    xmlHashAddEntry3(style->templatesHash, name, mode, modeURI, pat);
-#ifdef DEBUG_PARSING
-	    xsltGenericDebug(xsltGenericDebugContext,
-		    "xsltAddTemplate: added new hash %s\n", name);
-#endif
-	} else {
-	    list = (xsltCompMatchPtr) xmlHashLookup3(style->templatesHash,
-		                      name, mode, modeURI);
-	    if (list == NULL) {
-		xmlHashAddEntry3(style->templatesHash, name,
-			         mode, modeURI, pat);
+		style->templatesHash = xmlHashCreate(1024);
+		if (style->templatesHash == NULL) {
+		    xsltFreeCompMatch(pat);
+		    return(-1);
+		}
 #ifdef DEBUG_PARSING
 		xsltGenericDebug(xsltGenericDebugContext,
-			"xsltAddTemplate: added new hash %s\n", name);
+				 "xsltAddTemplate: created template hash\n");
+#endif
+		xmlHashAddEntry3(style->templatesHash, name, mode, modeURI, pat);
+#ifdef DEBUG_PARSING
+		xsltGenericDebug(xsltGenericDebugContext,
+				 "xsltAddTemplate: added new hash %s\n", name);
 #endif
 	    } else {
-		/*
-		 * Note '<=' since one must choose among the matching
-		 * template rules that are left, the one that occurs
-		 * last in the stylesheet
-		 */
-		if (list->priority <= pat->priority) {
-		    pat->next = list;
-		    xmlHashUpdateEntry3(style->templatesHash, name,
-			                mode, modeURI, pat, NULL);
+		list = (xsltCompMatchPtr) xmlHashLookup3(style->templatesHash,
+							 name, mode, modeURI);
+		if (list == NULL) {
+		    xmlHashAddEntry3(style->templatesHash, name,
+				     mode, modeURI, pat);
 #ifdef DEBUG_PARSING
 		    xsltGenericDebug(xsltGenericDebugContext,
-			    "xsltAddTemplate: added head hash for %s\n", name);
+				     "xsltAddTemplate: added new hash %s\n", name);
 #endif
 		} else {
-		    while (list->next != NULL) {
-			if (list->next->priority <= pat->priority)
-			    break;
-			list = list->next;
+		    /*
+		     * Note '<=' since one must choose among the matching
+		     * template rules that are left, the one that occurs
+		     * last in the stylesheet
+		     */
+		    if (list->priority <= pat->priority) {
+			pat->next = list;
+			xmlHashUpdateEntry3(style->templatesHash, name,
+					    mode, modeURI, pat, NULL);
+#ifdef DEBUG_PARSING
+			xsltGenericDebug(xsltGenericDebugContext,
+					 "xsltAddTemplate: added head hash for %s\n", name);
+#endif
+		    } else {
+			while (list->next != NULL) {
+			    if (list->next->priority <= pat->priority)
+				break;
+			    list = list->next;
+			}
+			pat->next = list->next;
+			list->next = pat;
 		    }
-		    pat->next = list->next;
-		    list->next = pat;
 		}
 	    }
-	}
-    } else if (top != NULL) {
-	list = *top;
-	if (list == NULL) {
-	    *top = pat;
-	    pat->next = NULL;
-	} else if (list->priority <= pat->priority) {
-	    pat->next = list;
-	    *top = pat;
-	} else {
-	    while (list->next != NULL) {
-		if (list->next->priority <= pat->priority)
-		    break;
-		list = list->next;
+	} else if (top != NULL) {
+	    list = *top;
+	    if (list == NULL) {
+		*top = pat;
+		pat->next = NULL;
+	    } else if (list->priority <= pat->priority) {
+		pat->next = list;
+		*top = pat;
+	    } else {
+		while (list->next != NULL) {
+		    if (list->next->priority <= pat->priority)
+			break;
+		    list = list->next;
+		}
+		pat->next = list->next;
+		list->next = pat;
 	    }
-	    pat->next = list->next;
-	    list->next = pat;
+	} else {
+	    xsltGenericError(xsltGenericErrorContext,
+			     "xsltAddTemplate: invalid compiled pattern\n");
+	    xsltFreeCompMatch(pat);
+	    return(-1);
 	}
-    } else {
-	xsltGenericError(xsltGenericErrorContext,
-		"xsltAddTemplate: invalid compiled pattern\n");
-	xsltFreeCompMatch(pat);
-	return(-1);
+	pat = next;
     }
-    if (*p != 0)
-	goto next_pattern;
     return(0);
 }
 
 /**
  * xsltGetTemplate:
  * @ctxt:  a XSLT process context
- * @mode:  the mode name or NULL
+ * @mode:  the mode 
+ * @style:  the current style
  *
- * Finds the template applying to this node
+ * Finds the template applying to this node, if @style is non-NULL
+ * it means one need to look for the next imported template in scope.
  *
  * Returns the xsltTemplatePtr or NULL if not found
  */
 xsltTemplatePtr
-xsltGetTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node) {
-    xsltStylesheetPtr style;
+xsltGetTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
+	        xsltStylesheetPtr style) {
+    xsltStylesheetPtr curstyle;
     xsltTemplatePtr ret = NULL;
     const xmlChar *name = NULL;
     xsltCompMatchPtr list = NULL;
@@ -1403,10 +1429,15 @@ xsltGetTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node) {
     if ((ctxt == NULL) || (node == NULL))
 	return(NULL);
 
-    style = ctxt->style;
-    while (style != NULL) {
+    if (style == NULL) {
+	curstyle = ctxt->style;
+    } else {
+	curstyle = xsltNextImport(style);
+    }
+
+    while ((curstyle != NULL) && (curstyle != style)) {
 	/* TODO : handle IDs/keys here ! */
-	if (style->templatesHash != NULL) {
+	if (curstyle->templatesHash != NULL) {
 	    /*
 	     * Use the top name as selector
 	     */
@@ -1443,7 +1474,7 @@ xsltGetTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 	    /*
 	     * find the list of appliable expressions based on the name
 	     */
-	    list = (xsltCompMatchPtr) xmlHashLookup3(style->templatesHash,
+	    list = (xsltCompMatchPtr) xmlHashLookup3(curstyle->templatesHash,
 					     name, ctxt->mode, ctxt->modeURI);
 	}
 	while (list != NULL) {
@@ -1461,24 +1492,24 @@ xsltGetTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 	 */
 	switch (node->type) {
 	    case XML_ELEMENT_NODE:
-		list = style->elemMatch;
+		list = curstyle->elemMatch;
 		break;
 	    case XML_ATTRIBUTE_NODE:
-		list = style->attrMatch;
+		list = curstyle->attrMatch;
 		break;
 	    case XML_PI_NODE:
-		list = style->piMatch;
+		list = curstyle->piMatch;
 		break;
 	    case XML_DOCUMENT_NODE:
 	    case XML_HTML_DOCUMENT_NODE:
-		list = style->rootMatch;
+		list = curstyle->rootMatch;
 		break;
 	    case XML_TEXT_NODE:
 	    case XML_CDATA_SECTION_NODE:
-		list = style->textMatch;
+		list = curstyle->textMatch;
 		break;
 	    case XML_COMMENT_NODE:
-		list = style->commentMatch;
+		list = curstyle->commentMatch;
 		break;
 	    case XML_ENTITY_REF_NODE:
 	    case XML_ENTITY_NODE:
@@ -1507,7 +1538,7 @@ xsltGetTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 	    list = list->next;
 	}
 	if (node->_private != NULL) {
-	    list = style->keyMatch;
+	    list = curstyle->keyMatch;
 	    while ((list != NULL) &&
 		   ((ret == NULL)  || (list->priority > ret->priority))) {
 		if (xsltTestCompMatch(ctxt, list, node,
@@ -1523,9 +1554,9 @@ xsltGetTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 	    return(ret);
 
 	/*
-	 * Cycle on next stylesheet import.
+	 * Cycle on next curstylesheet import.
 	 */
-	style = xsltNextImport(style);
+	curstyle = xsltNextImport(curstyle);
     }
     return(NULL);
 }
@@ -1560,3 +1591,31 @@ xsltFreeTemplateHashes(xsltStylesheetPtr style) {
         xsltFreeCompMatchList(style->commentMatch);
 }
 
+/**
+ * xsltMatchPattern
+ * @node: a node in the source tree
+ * @pattern: an XSLT pattern
+ *
+ * Determine if a node matches a pattern.
+ */
+int
+xsltMatchPattern(xsltTransformContextPtr context,
+		 xmlNodePtr node,
+		 const xmlChar *pattern)
+{
+    int match = 0;
+    xsltCompMatchPtr first, comp;
+
+    if ((context != NULL) && (pattern != NULL)) {
+	first = xsltCompilePattern(pattern);
+	for (comp = first; comp != NULL; comp = comp->next) {
+	    match = xsltTestCompMatch(context, comp, node, NULL, NULL);
+	    if (match)
+		break; /* for */
+	}
+	if (first)
+	    xsltFreeCompMatchList(first);
+    }
+    return match;
+}
+
