@@ -30,6 +30,7 @@
 
 #define DEBUG_TEMPLATES
 
+
 /************************************************************************
  *									*
  *			Module interfaces				*
@@ -163,7 +164,7 @@ xsltEvalTemplateString(xsltTransformContextPtr ctxt, xmlNodePtr node,
     oldInsert = ctxt->insert;
     ctxt->insert = insert;
 
-    xsltApplyOneTemplate(ctxt, node, parent->children);
+    xsltApplyOneTemplate(ctxt, node, NULL, parent->children);
 
     ctxt->insert = oldInsert;
 
@@ -291,7 +292,7 @@ xsltAttrTemplateProcess(xsltTransformContextPtr ctxt, xmlNodePtr target,
 	if (xmlStrEqual(cur->name, (const xmlChar *)"use-attribute-sets")) {
 	    xmlChar *in;
 
-	    in = xmlNodeListGetString(ctxt->doc, cur->children, 1);
+	    in = xmlNodeListGetString(ctxt->document->doc, cur->children, 1);
 	    if (in != NULL) {
 		xsltApplyAttributeSet(ctxt, ctxt->node, NULL, in);
 		xmlFree(in);
@@ -309,7 +310,8 @@ xsltAttrTemplateProcess(xsltTransformContextPtr ctxt, xmlNodePtr target,
 	ret->ns = NULL;
 
     if (cur->children != NULL) {
-	xmlChar *in = xmlNodeListGetString(ctxt->doc, cur->children, 1);
+	xmlChar *in = xmlNodeListGetString(ctxt->document->doc,
+		                           cur->children, 1);
 	xmlChar *out;
 
 	/* TODO: optimize if no template value was detected */
