@@ -22,6 +22,7 @@
 #include <libxml/encoding.h>
 #include <libxml/xmlerror.h>
 #include <libxml/xpath.h>
+#include <libxml/parserInternals.h>
 #include <libxml/xpathInternals.h>
 #include <libxml/HTMLtree.h>
 #include "xslt.h"
@@ -36,6 +37,7 @@
 #include "templates.h"
 #include "imports.h"
 #include "keys.h"
+#include "documents.h"
 
 #define DEBUG_PROCESS
 
@@ -51,6 +53,43 @@
 #define IS_BLANK_NODE(n)						\
     (((n)->type == XML_TEXT_NODE) && (xsltIsBlank((n)->content)))
 
+/*
+ * Generic function for accessing stacks in the transform Context
+ */
+
+#define PUSH_AND_POP(scope, type, name)					\
+scope int name##Push(xsltTransformContextPtr ctxt, type value) {	\
+    if (ctxt->name##Nr >= ctxt->name##Max) {				\
+	ctxt->name##Max *= 2;						\
+        ctxt->name##Tab = (type *) xmlRealloc(ctxt->name##Tab,		\
+	             ctxt->name##Max * sizeof(ctxt->name##Tab[0]));	\
+        if (ctxt->name##Tab == NULL) {					\
+	    xmlGenericError(xmlGenericErrorContext,			\
+		    "realloc failed !\n");				\
+	    return(0);							\
+	}								\
+    }									\
+    ctxt->name##Tab[ctxt->name##Nr] = value;				\
+    ctxt->name = value;							\
+    return(ctxt->name##Nr++);						\
+}									\
+scope type name##Pop(xsltTransformContextPtr ctxt) {			\
+    type ret;								\
+    if (ctxt->name##Nr <= 0) return(0);					\
+    ctxt->name##Nr--;							\
+    if (ctxt->name##Nr > 0)						\
+	ctxt->name = ctxt->name##Tab[ctxt->name##Nr - 1];		\
+    else								\
+        ctxt->name = NULL;						\
+    ret = ctxt->name##Tab[ctxt->name##Nr];				\
+    ctxt->name##Tab[ctxt->name##Nr] = 0;				\
+    return(ret);							\
+}									\
+
+/*
+ * Those macros actually generate the functions
+ */
+PUSH_AND_POP(extern, xsltTemplatePtr, templ)
 
 /************************************************************************
  *									*
@@ -60,6 +99,7 @@
 
 /**
  * xsltNewTransformContext:
+ * @style:  a parsed XSLT stylesheet
  * @doc:  the input document
  *
  * Create a new XSLT TransformContext
@@ -67,8 +107,9 @@
  * Returns the newly allocated xsltTransformContextPtr or NULL in case of error
  */
 xsltTransformContextPtr
-xsltNewTransformContext(xmlDocPtr doc) {
+xsltNewTransformContext(xsltStylesheetPtr style, xmlDocPtr doc) {
     xsltTransformContextPtr cur;
+    xsltDocumentPtr docu;
 
     cur = (xsltTransformContextPtr) xmlMalloc(sizeof(xsltTransformContext));
     if (cur == NULL) {
@@ -77,16 +118,38 @@ xsltNewTransformContext(xmlDocPtr doc) {
 	return(NULL);
     }
     memset(cur, 0, sizeof(xsltTransformContext));
+    cur->templTab = (xsltTemplatePtr *)
+	        xmlMalloc(10 * sizeof(xsltTemplatePtr));
+    if (cur->templTab == NULL) {
+        xmlGenericError(xmlGenericErrorContext,
+		"xsltNewTransformContext: out of memory\n");
+	xmlFree(cur);
+	return(NULL);
+    }
+    cur->templNr = 0;
+    cur->templMax = 5;
+    cur->templ = NULL;
+    cur->style = style;
     xmlXPathInit();
-    cur->doc = doc;
     cur->xpathCtxt = xmlXPathNewContext(doc);
     if (cur->xpathCtxt == NULL) {
         xsltGenericError(xsltGenericErrorContext,
 		"xsltNewTransformContext : xmlXPathNewContext failed\n");
+	xmlFree(cur->templTab);
 	xmlFree(cur);
 	return(NULL);
     }
     XSLT_REGISTER_VARIABLE_LOOKUP(cur);
+    docu = xsltNewDocument(cur, doc);
+    if (docu == NULL) {
+        xsltGenericError(xsltGenericErrorContext,
+		"xsltNewTransformContext : xsltNewDocument failed\n");
+	xmlFree(cur->templTab);
+	xmlFree(cur);
+	return(NULL);
+    }
+    docu->main = 1;
+    cur->document = docu;
     return(cur);
 }
 
@@ -98,20 +161,14 @@ xsltNewTransformContext(xmlDocPtr doc) {
  */
 void
 xsltFreeTransformContext(xsltTransformContextPtr ctxt) {
-    xmlDocPtr doc, next;
-
     if (ctxt == NULL)
 	return;
-    doc = ctxt->extraDocs;
-    while (doc != NULL) {
-	next = (xmlDocPtr) doc->next;
-	xmlFreeDoc(doc);
-	doc = next;
-    }
     if (ctxt->xpathCtxt != NULL)
 	xmlXPathFreeContext(ctxt->xpathCtxt);
+    if (ctxt->templTab != NULL)
+	xmlFree(ctxt->templTab);
     xsltFreeVariableHashes(ctxt);
-    xsltFreeCtxtKeys(ctxt);
+    xsltFreeDocuments(ctxt);
     memset(ctxt, -1, sizeof(xsltTransformContext));
     xmlFree(ctxt);
 }
@@ -445,7 +502,15 @@ xsltCopy(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	}
     }
 
-    xsltApplyOneTemplate(ctxt, ctxt->node, inst->children);
+    switch (node->type) {
+	case XML_DOCUMENT_NODE:
+	case XML_HTML_DOCUMENT_NODE:
+	case XML_ELEMENT_NODE:
+	    xsltApplyOneTemplate(ctxt, ctxt->node, NULL, inst->children);
+	    break;
+	default:
+	    break;
+    }
     ctxt->insert = oldInsert;
 }
 
@@ -524,7 +589,7 @@ xsltElement(xsltTransformContextPtr ctxt, xmlNodePtr node,
         xmlFree(attributes);
     }
     
-    xsltApplyOneTemplate(ctxt, ctxt->node, inst->children);
+    xsltApplyOneTemplate(ctxt, ctxt->node, NULL, inst->children);
 
 error:
     if (prop != NULL)
@@ -638,6 +703,12 @@ xsltCopyOf(xsltTransformContextPtr ctxt, xmlNodePtr node,
     if ((ctxt == NULL) || (node == NULL) || (inst == NULL))
 	return;
 
+    prop = xmlGetNsProp(inst, (const xmlChar *)"select", XSLT_NAMESPACE);
+    if (prop == NULL) {
+	xsltGenericError(xsltGenericErrorContext,
+	     "xslt:copy-of : select is missing\n");
+	goto error;
+    }
 #ifdef DEBUG_PROCESS
     xsltGenericDebug(xsltGenericDebugContext,
 	 "xsltCopyOf: select %s\n", prop);
@@ -748,9 +819,6 @@ xsltValueOf(xsltTransformContextPtr ctxt, xmlNodePtr node,
 		 "invalud value %s for disable-output-escaping\n", prop);
 
 	xmlFree(prop);
-	if (disableEscaping) {
-	    TODO /* disable-output-escaping */
-	}
     }
     prop = xmlGetNsProp(inst, (const xmlChar *)"select", XSLT_NAMESPACE);
     if (prop == NULL) {
@@ -781,6 +849,8 @@ xsltValueOf(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	if (res->type == XPATH_STRING) {
             copy = xmlNewText(res->stringval);
 	    if (copy != NULL) {
+		if (disableEscaping)
+		    copy->name = xmlStringTextNoenc;
 		xmlAddChild(ctxt->insert, copy);
 	    }
 	}
@@ -826,59 +896,31 @@ xsltNumber(xsltTransformContextPtr ctxt,
 
     memset(&numdata, 0, sizeof(numdata));
     
-    prop = xmlGetNsProp(cur, (const xmlChar *)"level", XSLT_NAMESPACE);
-    if (prop != NULL) {
-	if (xmlStrEqual(prop, BAD_CAST("single"))) {
-	    TODO;
-	} else if (xmlStrEqual(prop, BAD_CAST("multiple"))) {
-	    TODO;
-	} else if (xmlStrEqual(prop, BAD_CAST("any"))) {
-	    TODO;
-	} else {
-	    xsltGenericError(xsltGenericErrorContext,
-			     "invalid value %s for level\n", prop);
-	}
-	xmlFree(prop);
-    }
-    
-    prop = xmlGetNsProp(cur, (const xmlChar *)"count", XSLT_NAMESPACE);
-    if (prop != NULL) {
-	TODO;
-	xmlFree(prop);
-    }
-    
-    prop = xmlGetNsProp(cur, (const xmlChar *)"from", XSLT_NAMESPACE);
-    if (prop != NULL) {
-	TODO;
-	xmlFree(prop);
-    }
-    
-    prop = xmlGetNsProp(cur, (const xmlChar *)"value", XSLT_NAMESPACE);
-    if (prop != NULL) {
-	numdata.value = prop;
-    } else {
-	numdata.value = xmlStrdup(BAD_CAST("position()"));
-    }
+    numdata.value = xmlGetNsProp(cur, (const xmlChar *)"value", XSLT_NAMESPACE);
     
     prop = xmlGetNsProp(cur, (const xmlChar *)"format", XSLT_NAMESPACE);
     if (prop != NULL) {
-	/* Unicode categories:
-	 *  Nd = Number, decimal digit
-	 *  Nl = Number, letter
-	 *  No = Number, other
-	 *  Lu = Letters, uppercase
-	 *  Ll = Letters, lowercase
-	 *  Lt = Letters, titlecase
-	 *  Lm = Letters, modifiers
-	 *  Lo = Letters, other (uncased)
-	 *
-	 *  This corresponds to isalnum() in a Unicode locale.
-	 */
 	numdata.format = prop;
     } else {
 	numdata.format = xmlStrdup(BAD_CAST("1"));
     }
     
+    numdata.count = xmlGetNsProp(cur, (const xmlChar *)"count", XSLT_NAMESPACE);
+    numdata.from = xmlGetNsProp(cur, (const xmlChar *)"from", XSLT_NAMESPACE);
+    
+    prop = xmlGetNsProp(cur, (const xmlChar *)"level", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	if (xmlStrEqual(prop, BAD_CAST("single")) ||
+	    xmlStrEqual(prop, BAD_CAST("multiple")) ||
+	    xmlStrEqual(prop, BAD_CAST("any"))) {
+	    numdata.level = prop;
+	} else {
+	    xsltGenericError(xsltGenericErrorContext,
+			     "invalid value %s for level\n", prop);
+	    xmlFree(prop);
+	}
+    }
+    
     prop = xmlGetNsProp(cur, (const xmlChar *)"lang", XSLT_NAMESPACE);
     if (prop != NULL) {
 	TODO;
@@ -912,12 +954,25 @@ xsltNumber(xsltTransformContextPtr ctxt,
 	numdata.groupingCharacter = 0;
     }
 
+    /* Set default values */
+    if (numdata.value == NULL) {
+	if (numdata.level == NULL) {
+	    numdata.level = xmlStrdup(BAD_CAST("single"));
+	}
+    }
+    
     xsltNumberFormat(ctxt, &numdata, node);
 
-    if (numdata.format != NULL)
-	xmlFree(numdata.format);
+    if (numdata.level != NULL)
+	xmlFree(numdata.level);
+    if (numdata.count != NULL)
+	xmlFree(numdata.count);
+    if (numdata.from != NULL)
+	xmlFree(numdata.from);
     if (numdata.value != NULL)
 	xmlFree(numdata.value);
+    if (numdata.format != NULL)
+	xmlFree(numdata.format);
 }
 
 /**
@@ -960,16 +1015,59 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 	case XML_HTML_DOCUMENT_NODE:
 	case XML_ELEMENT_NODE:
 	    break;
-	case XML_TEXT_NODE:
-	    template = xsltGetTemplate(ctxt, node);
+	case XML_CDATA_SECTION_NODE:
+	    template = xsltGetTemplate(ctxt, node, NULL);
 	    if (template) {
 		xmlNodePtr oldNode;
 
+#ifdef DEBUG_PROCESS
+		xsltGenericDebug(xsltGenericDebugContext,
+		 "xsltDefaultProcessOneNode: applying template for CDATA %s\n",
+		                 node->content);
+#endif
 		oldNode = ctxt->node;
 		ctxt->node = node;
-		xsltApplyOneTemplate(ctxt, node, template->content);
+		xsltApplyOneTemplate(ctxt, node, template, NULL);
 		ctxt->node = oldNode;
-	    } else if (ctxt->mode == NULL) {
+	    } else /* if (ctxt->mode == NULL) */ {
+#ifdef DEBUG_PROCESS
+		xsltGenericDebug(xsltGenericDebugContext,
+		 "xsltDefaultProcessOneNode: copy CDATA %s\n",
+		                 node->content);
+#endif
+		copy = xmlNewDocText(ctxt->output, node->content);
+		if (copy != NULL) {
+		    xmlAddChild(ctxt->insert, copy);
+		} else {
+		    xsltGenericError(xsltGenericErrorContext,
+			"xsltDefaultProcessOneNode: cdata copy failed\n");
+		}
+	    }
+	    return;
+	case XML_TEXT_NODE:
+	    template = xsltGetTemplate(ctxt, node, NULL);
+	    if (template) {
+		xmlNodePtr oldNode;
+
+#ifdef DEBUG_PROCESS
+		xsltGenericDebug(xsltGenericDebugContext,
+		 "xsltDefaultProcessOneNode: applying template for text %s\n",
+		                 node->content);
+#endif
+		oldNode = ctxt->node;
+		ctxt->node = node;
+		xsltApplyOneTemplate(ctxt, node, template, NULL);
+		ctxt->node = oldNode;
+	    } else /* if (ctxt->mode == NULL) */ {
+#ifdef DEBUG_PROCESS
+		if (node->content == NULL)
+		    xsltGenericDebug(xsltGenericDebugContext,
+		     "xsltDefaultProcessOneNode: copy empty text\n");
+		else
+		    xsltGenericDebug(xsltGenericDebugContext,
+		     "xsltDefaultProcessOneNode: copy text %s\n",
+				     node->content);
+#endif
 		copy = xmlCopyNode(node, 0);
 		if (copy != NULL) {
 		    xmlAddChild(ctxt->insert, copy);
@@ -982,13 +1080,13 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 	case XML_ATTRIBUTE_NODE:
 	    if (ctxt->insert->type == XML_ELEMENT_NODE) {
 		    xmlAttrPtr attr = (xmlAttrPtr) node, ret = NULL, cur;
-		template = xsltGetTemplate(ctxt, node);
+		template = xsltGetTemplate(ctxt, node, NULL);
 		if (template) {
 		    xmlNodePtr oldNode;
 
 		    oldNode = ctxt->node;
 		    ctxt->node = node;
-		    xsltApplyOneTemplate(ctxt, node, template->content);
+		    xsltApplyOneTemplate(ctxt, node, template, NULL);
 		    ctxt->node = oldNode;
 		} else if (ctxt->mode == NULL) {
 		    if (attr->ns != NULL) {
@@ -1069,13 +1167,13 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
      */
     attrs = node->properties;
     while (attrs != NULL) {
-	template = xsltGetTemplate(ctxt, (xmlNodePtr) attrs);
+	template = xsltGetTemplate(ctxt, (xmlNodePtr) attrs, NULL);
 	if (template) {
 	    xmlNodePtr oldNode;
 
 	    oldNode = ctxt->node;
 	    ctxt->node = node;
-	    xsltApplyOneTemplate(ctxt, node, template->content);
+	    xsltApplyOneTemplate(ctxt, node, template, NULL);
 	    ctxt->node = oldNode;
 	}
 	attrs = attrs->next;
@@ -1093,19 +1191,61 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 		ctxt->xpathCtxt->proximityPosition = childno;
 		xsltProcessOneNode(ctxt, cur);
 		break;
-	    case XML_TEXT_NODE:
 	    case XML_CDATA_SECTION_NODE:
-		template = xsltGetTemplate(ctxt, cur);
+		template = xsltGetTemplate(ctxt, node, NULL);
 		if (template) {
 		    xmlNodePtr oldNode;
 
+#ifdef DEBUG_PROCESS
+		    xsltGenericDebug(xsltGenericDebugContext,
+		 "xsltDefaultProcessOneNode: applying template for CDATA %s\n",
+				     node->content);
+#endif
+		    oldNode = ctxt->node;
+		    ctxt->node = node;
+		    xsltApplyOneTemplate(ctxt, node, template, NULL);
+		    ctxt->node = oldNode;
+		} else /* if (ctxt->mode == NULL) */ {
+#ifdef DEBUG_PROCESS
+		    xsltGenericDebug(xsltGenericDebugContext,
+		     "xsltDefaultProcessOneNode: copy CDATA %s\n",
+				     node->content);
+#endif
+		    copy = xmlNewDocText(ctxt->output, node->content);
+		    if (copy != NULL) {
+			xmlAddChild(ctxt->insert, copy);
+		    } else {
+			xsltGenericError(xsltGenericErrorContext,
+			    "xsltDefaultProcessOneNode: cdata copy failed\n");
+		    }
+		}
+		break;
+	    case XML_TEXT_NODE:
+		template = xsltGetTemplate(ctxt, cur, NULL);
+		if (template) {
+		    xmlNodePtr oldNode;
+
+#ifdef DEBUG_PROCESS
+		    xsltGenericDebug(xsltGenericDebugContext,
+	     "xsltDefaultProcessOneNode: applying template for text %s\n",
+				     node->content);
+#endif
 		    oldNode = ctxt->node;
 		    ctxt->node = cur;
 		    ctxt->xpathCtxt->contextSize = nbchild;
 		    ctxt->xpathCtxt->proximityPosition = childno;
-		    xsltApplyOneTemplate(ctxt, cur, template->content);
+		    xsltApplyOneTemplate(ctxt, cur, template, NULL);
 		    ctxt->node = oldNode;
-		} else if (ctxt->mode == NULL) {
+		} else /* if (ctxt->mode == NULL) */ {
+#ifdef DEBUG_PROCESS
+		    if (cur->content == NULL)
+			xsltGenericDebug(xsltGenericDebugContext,
+			 "xsltDefaultProcessOneNode: copy empty text\n");
+		    else
+			xsltGenericDebug(xsltGenericDebugContext,
+		     "xsltDefaultProcessOneNode: copy text %s\n",
+					 cur->content);
+#endif
 		    copy = xmlCopyNode(cur, 0);
 		    if (copy != NULL) {
 			xmlAddChild(ctxt->insert, copy);
@@ -1117,7 +1257,7 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 		break;
 	    case XML_PI_NODE:
 	    case XML_COMMENT_NODE:
-		template = xsltGetTemplate(ctxt, cur);
+		template = xsltGetTemplate(ctxt, cur, NULL);
 		if (template) {
 		    xmlNodePtr oldNode;
 
@@ -1125,7 +1265,7 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 		    ctxt->node = cur;
 		    ctxt->xpathCtxt->contextSize = nbchild;
 		    ctxt->xpathCtxt->proximityPosition = childno;
-		    xsltApplyOneTemplate(ctxt, cur, template->content);
+		    xsltApplyOneTemplate(ctxt, cur, template, NULL);
 		    ctxt->node = oldNode;
 		}
 		break;
@@ -1138,6 +1278,30 @@ xsltDefaultProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
     ctxt->xpathCtxt->proximityPosition = oldPos;
 }
 
+/**
+ * xsltApplyImports:
+ * @ctxt:  a XSLT process context
+ * @node:  the node in the source tree.
+ * @inst:  the xslt apply-imports node
+ *
+ * Process the xslt apply-imports node on the source node
+ */
+void
+xsltApplyImports(xsltTransformContextPtr ctxt, xmlNodePtr node,
+	         xmlNodePtr inst) {
+    xsltTemplatePtr template;
+
+    if ((ctxt->templ == NULL) || (ctxt->templ->style == NULL)) {
+	xsltGenericError(xsltGenericErrorContext,
+	     "xslt:apply-imports : internal error no current template\n");
+	return;
+    }
+    template = xsltGetTemplate(ctxt, node, ctxt->templ->style);
+    if (template != NULL) {
+	xsltApplyOneTemplate(ctxt, node, template, NULL);
+    }
+}
+
 /**
  * xsltCallTemplate:
  * @ctxt:  a XSLT process context
@@ -1209,7 +1373,7 @@ xsltCallTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	}
 	cur = cur->next;
     }
-    xsltApplyOneTemplate(ctxt, node, template->content);
+    xsltApplyOneTemplate(ctxt, node, template, NULL);
 
 error:
     if (has_param == 1)
@@ -1438,23 +1602,35 @@ error:
  * xsltApplyOneTemplate:
  * @ctxt:  a XSLT process context
  * @node:  the node in the source tree.
+ * @templ:  the template
  * @list:  the template replacement nodelist
  *
  * Process the apply-templates node on the source node
  */
 void
 xsltApplyOneTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
-	             xmlNodePtr list) {
+	             xsltTemplatePtr templ, xmlNodePtr list) {
     xmlNodePtr cur = NULL, insert, copy = NULL;
     xmlNodePtr oldInsert;
     xmlAttrPtr attrs;
     int has_variables = 0;
 
+    if ((templ == NULL) && (list == NULL))
+	return;
     CHECK_STOPPED;
+
+    /*
+     * stack and saves
+     */
+    if (templ != NULL)
+	templPush(ctxt, templ);
     oldInsert = insert = ctxt->insert;
+
     /*
      * Insert all non-XSLT nodes found in the template
      */
+    if ((list == NULL) && (templ != NULL))
+	list = templ->content;
     cur = list;
     while (cur != NULL) {
 	/*
@@ -1465,6 +1641,11 @@ xsltApplyOneTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	    xsltGenericDebug(xsltGenericDebugContext,
 		 "xsltApplyOneTemplate: insert == NULL !\n");
 #endif
+	    if (has_variables != 0) {
+		xsltPopStack(ctxt);
+	    }
+	    if (templ != NULL)
+		templPop(ctxt);
 	    return;
 	}
 
@@ -1497,6 +1678,10 @@ xsltApplyOneTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
 		ctxt->insert = insert;
 		xsltForEach(ctxt, node, cur);
 		ctxt->insert = oldInsert;
+	    } else if (IS_XSLT_NAME(cur, "apply-imports")) {
+		ctxt->insert = insert;
+		xsltApplyImports(ctxt, node, cur);
+		ctxt->insert = oldInsert;
 	    } else if (IS_XSLT_NAME(cur, "attribute")) {
 		ctxt->insert = insert;
 		xsltAttribute(ctxt, node, cur);
@@ -1555,8 +1740,13 @@ xsltApplyOneTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	     * element names consists of just xsl:text.
 	     */
 #ifdef DEBUG_PROCESS
-	    xsltGenericDebug(xsltGenericDebugContext,
-		 "xsltApplyOneTemplate: copy text %s\n", cur->content);
+	    if (cur->name == xmlStringTextNoenc)
+		xsltGenericDebug(xsltGenericDebugContext,
+		     "xsltApplyOneTemplate: copy unescaped text %s\n",
+		                 cur->content);
+	    else
+		xsltGenericDebug(xsltGenericDebugContext,
+		     "xsltApplyOneTemplate: copy text %s\n", cur->content);
 #endif
 	    copy = xmlCopyNode(cur, 0);
 	    if (copy != NULL) {
@@ -1565,6 +1755,9 @@ xsltApplyOneTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node,
 		xsltGenericError(xsltGenericErrorContext,
 			"xsltApplyOneTemplate: text copy failed\n");
 	    }
+	} else if ((cur->type == XML_ELEMENT_NODE) &&
+		   (xmlStrEqual(cur->name, (const xmlChar *)"xsltdebug"))) {
+	    xsltDebug(ctxt, cur);
 	} else if (cur->type == XML_ELEMENT_NODE) {
 #ifdef DEBUG_PROCESS
 	    xsltGenericDebug(xsltGenericDebugContext,
@@ -1622,6 +1815,8 @@ skip_children:
     if (has_variables != 0) {
 	xsltPopStack(ctxt);
     }
+    if (templ != NULL)
+	templPop(ctxt);
 }
 
 /**
@@ -1704,7 +1899,7 @@ xsltChoose(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	    "xsl:when: test evaluate to %d\n", doit);
 #endif
 	if (doit) {
-	    xsltApplyOneTemplate(ctxt, ctxt->node, when->children);
+	    xsltApplyOneTemplate(ctxt, ctxt->node, NULL, when->children);
 	    goto done;
 	}
 	if (xpathParserCtxt != NULL)
@@ -1719,7 +1914,7 @@ xsltChoose(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	replacement = replacement->next;
     }
     if (IS_XSLT_ELEM(replacement) && (IS_XSLT_NAME(replacement, "otherwise"))) {
-	xsltApplyOneTemplate(ctxt, ctxt->node, replacement->children);
+	xsltApplyOneTemplate(ctxt, ctxt->node, NULL, replacement->children);
 	replacement = replacement->next;
     }
     if (replacement != NULL) {
@@ -1803,7 +1998,7 @@ xsltIf(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	"xsltIf: test evaluate to %d\n", doit);
 #endif
     if (doit) {
-	xsltApplyOneTemplate(ctxt, node, inst->children);
+	xsltApplyOneTemplate(ctxt, node, NULL, inst->children);
     }
 
 error:
@@ -1895,7 +2090,7 @@ xsltForEach(xsltTransformContextPtr ctxt, xmlNodePtr node,
     for (i = 0;i < list->nodeNr;i++) {
 	ctxt->node = list->nodeTab[i];
 	ctxt->xpathCtxt->proximityPosition = i + 1;
-	xsltApplyOneTemplate(ctxt, list->nodeTab[i], replacement);
+	xsltApplyOneTemplate(ctxt, list->nodeTab[i], NULL, replacement);
     }
     ctxt->nodeList = oldlist;
     ctxt->xpathCtxt->contextSize = oldContextSize;
@@ -1922,7 +2117,7 @@ xsltProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
     xsltTemplatePtr template;
     xmlNodePtr oldNode;
 
-    template = xsltGetTemplate(ctxt, node);
+    template = xsltGetTemplate(ctxt, node, NULL);
     /*
      * If no template is found, apply the default rule.
      */
@@ -1931,6 +2126,9 @@ xsltProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 	if (node->type == XML_DOCUMENT_NODE)
 	    xsltGenericDebug(xsltGenericDebugContext,
 	     "xsltProcessOneNode: no template found for /\n");
+	else if (node->type == XML_CDATA_SECTION_NODE)
+	    xsltGenericDebug(xsltGenericDebugContext,
+	     "xsltProcessOneNode: no template found for CDATA\n");
 	else 
 	    xsltGenericDebug(xsltGenericDebugContext,
 	     "xsltProcessOneNode: no template found for %s\n", node->name);
@@ -1948,7 +2146,7 @@ xsltProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 	     "xsltProcessOneNode: applying template for attribute %s\n",
 	                 node->name);
 #endif
-	xsltApplyOneTemplate(ctxt, node, template->content);
+	xsltApplyOneTemplate(ctxt, node, template, NULL);
     } else {
 #ifdef DEBUG_PROCESS
 	if (node->type == XML_DOCUMENT_NODE)
@@ -1960,7 +2158,7 @@ xsltProcessOneNode(xsltTransformContextPtr ctxt, xmlNodePtr node) {
 #endif
 	oldNode = ctxt->node;
 	ctxt->node = node;
-	xsltApplyOneTemplate(ctxt, node, template->content);
+	xsltApplyOneTemplate(ctxt, node, template, NULL);
 	ctxt->node = oldNode;
     }
 }
@@ -1983,11 +2181,9 @@ xsltApplyStylesheet(xsltStylesheetPtr style, xmlDocPtr doc) {
 
     if ((style == NULL) || (doc == NULL))
 	return(NULL);
-    ctxt = xsltNewTransformContext(doc);
+    ctxt = xsltNewTransformContext(style, doc);
     if (ctxt == NULL)
 	return(NULL);
-    ctxt->style = style;
-    xsltInitCtxtKeys(ctxt);
     xsltEvalGlobalVariables(ctxt);
     if ((style->method != NULL) &&
 	(!xmlStrEqual(style->method, (const xmlChar *) "xml"))) {
@@ -1996,6 +2192,14 @@ xsltApplyStylesheet(xsltStylesheetPtr style, xmlDocPtr doc) {
 	    res = htmlNewDoc(style->doctypePublic, style->doctypeSystem);
 	    if (res == NULL)
 		goto error;
+	} else if (xmlStrEqual(style->method, (const xmlChar *) "xhtml")) {
+	    xsltGenericError(xsltGenericErrorContext,
+	     "xsltApplyStylesheet: insupported method xhtml, using html\n",
+		             style->method);
+	    ctxt->type = XSLT_OUTPUT_HTML;
+	    res = htmlNewDoc(style->doctypePublic, style->doctypeSystem);
+	    if (res == NULL)
+		goto error;
 	} else if (xmlStrEqual(style->method, (const xmlChar *) "text")) {
 	    ctxt->type = XSLT_OUTPUT_TEXT;
 	    res = xmlNewDoc(style->version);
