@@ -30,6 +30,7 @@
 #include "xsltutils.h"
 #include "imports.h"
 #include "keys.h"
+#include "documents.h"
 
 #define DEBUG_PARSING
 /* #define DEBUG_BLANKS */
@@ -287,6 +288,7 @@ xsltFreeStylesheet(xsltStylesheetPtr sheet) {
     xsltFreeTemplateList(sheet->templates);
     xsltFreeAttributeSetsHashes(sheet);
     xsltFreeNamespaceAliasHashes(sheet);
+    xsltFreeStyleDocuments(sheet);
     if (sheet->doc != NULL)
 	xmlFreeDoc(sheet->doc);
     if (sheet->variables != NULL)
@@ -1017,6 +1019,7 @@ xsltParseStylesheetTemplate(xsltStylesheetPtr style, xmlNodePtr template) {
 	return;
     ret->next = style->templates;
     style->templates = ret;
+    ret->style = style;
 
     /*
      * Get arguments
