@@ -52,6 +52,7 @@ typedef struct _xsltTemplate xsltTemplate;
 typedef xsltTemplate *xsltTemplatePtr;
 struct _xsltTemplate {
     struct _xsltTemplate *next;/* chained list sorted by priority */
+    struct _xsltStylesheet *style;/* the containing stylesheet */
     xmlChar *match;	/* the matching string */
     int priority;	/* as given from the stylesheet, not computed */
     xmlChar *name;	/* the local part of the name QName */
@@ -82,6 +83,19 @@ typedef struct _xsltDecimalFormat {
     xmlChar *zeroDigit;
 } xsltDecimalFormat, *xsltDecimalFormatPtr;
 
+/*
+ * Data structure associated to a document
+ */
+
+typedef struct _xsltDocument xsltDocument;
+typedef xsltDocument *xsltDocumentPtr;
+struct _xsltDocument {
+    struct _xsltDocument *next;	/* documents are kept in a chained list */
+    int main;			/* is this the main document */
+    xmlDocPtr doc;		/* the parsed document */
+    void *keys;			/* key tables storage */
+};
+
 /*
  * The in-memory structure corresponding to an XSLT Stylesheet
  * NOTE: most of the content is simply linked from the doc tree
@@ -97,6 +111,8 @@ struct _xsltStylesheet {
     struct _xsltStylesheet *next;
     struct _xsltStylesheet *imports;
 
+    xsltDocumentPtr docList;		/* the include document list */
+
     /*
      * General data on the style sheet document
      */
@@ -179,10 +195,17 @@ struct _xsltTransformContext {
     xsltStylesheetPtr style;		/* the stylesheet used */
     xsltOutputType type;		/* the type of output */
 
+    xsltTemplatePtr  templ;		/* the current template */
+    int              templNr;		/* Nb of templates in the stack */
+    int              templMax;		/* Size of the templtes stack */
+    xsltTemplatePtr *templTab;		/* the template stack */
+
     const xmlChar *mode;		/* the current mode */
     const xmlChar *modeURI;		/* the current mode URI */
 
-    xmlDocPtr doc;			/* the current doc */
+    xsltDocumentPtr docList;		/* the document list */
+
+    xsltDocumentPtr document;		/* the current document */
     xmlNodePtr node;			/* the current node */
     xmlNodeSetPtr nodeList;		/* the current node list */
 
@@ -192,8 +215,6 @@ struct _xsltTransformContext {
     xmlXPathContextPtr xpathCtxt;	/* the XPath context */
     void *variablesHash;		/* hash table or wherever variables
 				   	   informations are stored */
-    void *keys;				/* key tables storage */
-    xmlDocPtr extraDocs;		/* extra docs parsed by document() */
     xsltTransformState state;		/* the current state */
 };
 
